#
# parakeet.io.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import h5py
import numpy as np
import mrcfile
import os
import PIL.Image

try:
    FEI_EXTENDED_HEADER_DTYPE = mrcfile.dtypes.FEI1_EXTENDED_HEADER_DTYPE
except Exception:
    FEI_EXTENDED_HEADER_DTYPE = mrcfile.dtypes.FEI_EXTENDED_HEADER_DTYPE


class Writer(object):
    """
    Interface to write the simulated data

    """

    @property
    def shape(self):
        """
        The shape property

        """
        return self._data.shape

    @property
    def dtype(self):
        """
        The dtype property

        """
        return self._data.dtype

    @property
    def data(self):
        """
        The data property

        """
        return self._data

    @property
    def angle(self):
        """
        The angle property

        """
        return self._angle

    @property
    def position(self):
        """
        The position property

        """
        return self._position

    @property
    def drift(self):
        """
        The drift property

        """
        return self._drift

    @property
    def pixel_size(self):
        """
        The pixel size property

        """
        return self._pixel_size

    @property
    def defocus(self):
        """
        The defocus property

        """
        return self._defocus

    @property
    def is_mrcfile_writer(self):
        """
        Return if is mrcfile

        """
        return isinstance(self, MrcFileWriter)

    @property
    def is_nexus_writer(self):
        """
        Return if is nexus

        """
        return isinstance(self, NexusWriter)

    @property
    def is_image_writer(self):
        """
        Return if is image

        """
        return isinstance(self, ImageWriter)

    def update(self):
        """
        Update if anything needs to be done

        """
        pass


class MrcFileWriter(Writer):
    """
    Write to an mrcfile

    """

    class AngleProxy(object):
        """
        Proxy interface to angles

        """

        def __init__(self, handle):
            self.handle = handle

        def __setitem__(self, item, data):
            self.handle.extended_header[item]["Alpha tilt"] = data

    class PositionProxy(object):
        """
        Proxy interface to positions

        """

        def __init__(self, handle):
            self.handle = handle
            n = len(self.handle.extended_header)
            self.x, self.y = np.meshgrid(np.arange(0, 3), np.arange(0, n))

        def __setitem__(self, item, data):

            # Set the items
            def setitem_internal(j, i, d):
                if i == 0:
                    self.handle.extended_header[j]["Shift X"] = d
                elif i == 1:
                    self.handle.extended_header[j]["Shift Y"] = d
                elif i == 2:
                    self.handle.extended_header[j]["Z-Stage"] = d

            # Get the indices from the item
            x = self.x[item]
            y = self.y[item]

            # Set the item
            if isinstance(x, np.ndarray):
                for j, i, d in zip(y, x, data):
                    setitem_internal(j, i, d)
            else:
                setitem_internal(y, x, data)

    class DriftProxy(object):
        """
        Proxy interface to drifts

        """

        def __init__(self, handle):
            self.handle = handle
            n = len(self.handle.extended_header)
            self.x, self.y = np.meshgrid(np.arange(0, 2), np.arange(0, n))

        def __setitem__(self, item, data):

            # Set the items
            def setitem_internal(j, i, d):
                if i == 0:
                    self.handle.extended_header[j]["Shift offset X"] = d
                elif i == 1:
                    self.handle.extended_header[j]["Shift offset Y"] = d

            # Get the indices from the item
            x = self.x[item]
            y = self.y[item]

            # Set the item
            if isinstance(x, np.ndarray):
                for j, i, d in zip(y, x, data):
                    setitem_internal(j, i, d)
            else:
                setitem_internal(y, x, data)

    class DefocusProxy(object):
        """
        Proxy interface to angles

        """

        def __init__(self, handle):
            self.handle = handle

        def __setitem__(self, item, data):
            self.handle.extended_header[item]["Defocus"] = data

    def __init__(self, filename, shape, pixel_size, dtype="uint8"):
        """
        Initialise the writer

        Args:
            filename (str): The filename
            shape (tuple): The shape of the data
            pixel_size (float): The pixel size
            dtype (str): The data type

        """
        # Get the dtype
        dtype = np.dtype(dtype)

        # Convert 32bit int and 64bit float
        if dtype == "int32":
            dtype = np.dtype(np.int16)
        elif dtype == "uint32":
            dtype = np.dtype(np.uint16)
        elif dtype == "float64":
            dtype = np.dtype(np.float32)
        elif dtype == "complex128":
            dtype = np.dtype(np.complex64)

        # Open the handle to the mrcfile
        self.handle = mrcfile.new_mmap(
            filename,
            shape=(0, 0, 0),
            mrc_mode=mrcfile.utils.mode_from_dtype(dtype),
            overwrite=True,
        )

        # Setup the extended header
        extended_header = np.zeros(shape=shape[0], dtype=FEI_EXTENDED_HEADER_DTYPE)

        # Set the extended header
        self.handle._check_writeable()
        self.handle._close_data()
        self.handle._extended_header = extended_header
        self.handle.header.nsymbt = extended_header.nbytes
        self.handle.header.exttyp = "FEI1"
        self.handle._open_memmap(dtype, shape)
        self.handle.update_header_from_data()
        self.handle.flush()

        # Set the pixel size
        self.handle.voxel_size = pixel_size
        for i in range(self.handle.extended_header.shape[0]):
            self.handle.extended_header[i]["Pixel size X"] = pixel_size * 1e-10
            self.handle.extended_header[i]["Pixel size Y"] = pixel_size * 1e-10
            self.handle.extended_header[i]["Application"] = "RFI Simulation"

        # Set the data array
        self._data = self.handle.data
        self._angle = MrcFileWriter.AngleProxy(self.handle)
        self._position = MrcFileWriter.PositionProxy(self.handle)
        self._drift = MrcFileWriter.DriftProxy(self.handle)
        self._defocus = MrcFileWriter.DefocusProxy(self.handle)

    @property
    def pixel_size(self):
        """
        The pixel size

        """
        return self.handle.voxel_size[0]

    def update(self):
        """
        Update before closing

        """
        self.handle.update_header_stats()


class NexusWriter(Writer):
    """
    Write to a nexus file

    """

    class PositionProxy(object):
        """
        Proxy interface to positions

        """

        def __init__(self, handle):
            self.handle = handle
            n = self.handle["x_translation"].shape[0]
            self.x, self.y = np.meshgrid(np.arange(0, 3), np.arange(0, n))

        def __setitem__(self, item, data):

            # Set the items
            def setitem_internal(j, i, d):
                if i == 0:
                    self.handle["x_translation"][j] = d
                elif i == 1:
                    self.handle["y_translation"][j] = d
                elif i == 2:
                    self.handle["z_translation"][j] = d

            # Get the indices from the item
            x = self.x[item]
            y = self.y[item]

            # Set the item
            if isinstance(x, np.ndarray):
                for j, i, d in zip(y, x, data):
                    setitem_internal(j, i, d)
            else:
                setitem_internal(y, x, data)

    class ShiftProxy(object):
        """
        Proxy interface to shifts

        """

        def __init__(self, handle):
            self.handle = handle
            n = self.handle["x_drift"].shape[0]
            self.x, self.y = np.meshgrid(np.arange(0, 2), np.arange(0, n))

        def __setitem__(self, item, data):

            # Set the items
            def setitem_internal(j, i, d):
                if i == 0:
                    self.handle["x_drift"][j] = d
                elif i == 1:
                    self.handle["y_drift"][j] = d

            # Get the indices from the item
            x = self.x[item]
            y = self.y[item]

            # Set the item
            if isinstance(x, np.ndarray):
                for j, i, d in zip(y, x, data):
                    setitem_internal(j, i, d)
            else:
                setitem_internal(y, x, data)

    def __init__(self, filename, shape, pixel_size, dtype="float32"):
        """
        Initialise the writer

        Args:
            filename (str): The filename
            shape (tuple): The shape of the data
            pixel_size (float): The pixel size
            dtype (object): The data type of the data

        """

        # Open the file for writing
        self.handle = h5py.File(filename, "w")

        # Create the entry
        entry = self.handle.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        entry["definition"] = "NXtomo"

        # Create the instrument
        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"

        # Create the detector
        detector = instrument.create_group("detector")
        detector.attrs["NX_class"] = "NXdetector"
        detector.create_dataset("data", shape=shape, dtype=dtype)
        detector["image_key"] = np.zeros(shape=shape[0])
        detector["x_pixel_size"] = np.full(shape=shape[0], fill_value=pixel_size)
        detector["y_pixel_size"] = np.full(shape=shape[0], fill_value=pixel_size)

        # Create the sample
        sample = entry.create_group("sample")
        sample.attrs["NX_class"] = "NXsample"
        sample["name"] = "parakeet-simulation"
        sample.create_dataset("rotation_angle", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("x_translation", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("y_translation", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("z_translation", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("x_drift", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("y_drift", shape=(shape[0],), dtype=np.float32)
        sample.create_dataset("defocus", shape=(shape[0],), dtype=np.float32)

        # Create the data
        data = entry.create_group("data")
        data["data"] = detector["data"]
        data["rotation_angle"] = sample["rotation_angle"]
        data["x_translation"] = sample["x_translation"]
        data["y_translation"] = sample["y_translation"]
        data["z_translation"] = sample["z_translation"]
        data["x_drift"] = sample["x_drift"]
        data["y_drift"] = sample["y_drift"]
        data["defocus"] = sample["defocus"]
        data["image_key"] = detector["image_key"]

        # Set the data ptr
        self._data = data["data"]
        self._angle = data["rotation_angle"]
        self._position = NexusWriter.PositionProxy(data)
        self._drift = NexusWriter.ShiftProxy(data)
        self._defocus = data["defocus"]

    @property
    def pixel_size(self):
        """
        Return the pixel size

        """
        return self.handle["instrument"]["detector"]["x_pixel_size"][0]


class ImageWriter(Writer):
    """
    Write to a images

    """

    class DataProxy(object):
        """
        A proxy interface for the data

        """

        def __init__(self, template, shape=None, vmin=None, vmax=None):
            self.template = template
            self.shape = shape
            self.vmin = vmin
            self.vmax = vmax

        def __setitem__(self, item, data):

            # Check the input
            assert isinstance(item, tuple)
            assert isinstance(item[1], slice)
            assert isinstance(item[2], slice)
            assert item[1].start is None
            assert item[1].stop is None
            assert item[1].step is None
            assert item[2].start is None
            assert item[2].stop is None
            assert item[2].step is None
            assert len(data.shape) == 2
            assert data.shape[0] == self.shape[1]
            assert data.shape[1] == self.shape[2]

            # Convert to squared amplitude
            if np.iscomplexobj(data):
                data = np.abs(data) ** 2
                self.vmin = None
                self.vmax = None

            # Compute scale factors to put between 0 and 255
            if self.vmin is None:
                vmin = np.min(data)
            else:
                vmin = self.vmin
            if self.vmax is None:
                vmax = np.max(data)
            else:
                vmax = self.vmax
            s1 = 255.0 / (vmax - vmin)
            s0 = -s1 * vmin

            # Save the image to file
            filename = self.template % (item[0] + 1)
            image = (data * s1 + s0).astype(np.uint8)
            PIL.Image.fromarray(image).save(filename)

    def __init__(self, template, shape=None, vmin=None, vmax=None):
        """
        Initialise the writer

        Args:
            filename (str): The filename
            shape (tuple): The shape of the data

        """

        # Set the proxy data interface
        self._data = ImageWriter.DataProxy(template, shape, vmin, vmax)

        # Create dummy arrays for angle and position
        self._angle = np.zeros(shape=shape[0], dtype=np.float32)
        self._position = np.zeros(shape=(shape[0], 3), dtype=np.float32)
        self._drift = np.zeros(shape=(shape[0], 2), dtype=np.float32)
        self._defocus = np.zeros(shape=shape[0], dtype=np.float32)
        self._pixel_size = 0

    @property
    def vmin(self):
        """
        The vmin property

        """
        return self._data.vmin

    @property
    def vmax(self):
        """
        The vmax property

        """
        return self._data.vmax

    @vmin.setter
    def vmin(self, vmin):
        """
        The vmin propety setter

        """
        self._data.vmin = vmin

    @vmax.setter
    def vmax(self, vmax):
        """
        The vmax property setter

        """
        self._data.vmax = vmax


class Reader(object):
    """
    Interface to write the simulated data

    """

    def __init__(
        self, handle, data, angle, position, pixel_size, drift=None, defocus=None
    ):
        """
        Initialise the data

        Args:
            data (array): The data array
            angle (array): The angle array
            position (array): The position array
            pixel_size (float): The pixel size array
            drift (float): The drift array
            defocus (float): The defocus array

        """
        # Check the size
        assert len(angle) == data.shape[0], "Inconsistent dimensions"
        assert len(position) == data.shape[0], "Inconsistent dimensions"

        # Set the array
        self.handle = handle
        self.data = data
        self.angle = angle
        self.position = position
        self.pixel_size = pixel_size
        self.drift = drift
        self.defocus = defocus
        self.shape = data.shape
        self.dtype = data.dtype

    @property
    def start_angle(self):
        """
        Returns:
            float: the start angle

        """
        return self.angle[0]

    @property
    def stop_angle(self):
        """
        Returns:
            float: The stop angle

        """
        return self.angle[-1]

    @property
    def step_angle(self):
        """
        Returns:
            float: The stop angle

        """
        # tol = 1e-7
        if len(self.angle) == 1:
            step = 1
        else:
            step = (self.angle[-1] - self.angle[0]) / (len(self.angle) - 1)
        return step

    @property
    def num_images(self):
        """
        Returns:
            int: The number of images

        """
        return len(self.angle)

    @classmethod
    def from_mrcfile(Class, filename):
        """
        Read the simulated data from a mrc file

        Args:
            filename (str): The input filename

        """

        # Read the data
        handle = mrcfile.mmap(filename, "r")

        # Check the header info
        if handle.header.exttyp == b"FEI1":
            assert handle.extended_header.dtype == FEI_EXTENDED_HEADER_DTYPE
            assert len(handle.extended_header.shape) == 1
            assert handle.extended_header.shape[0] == handle.data.shape[0]

            # Read the angles
            angle = np.zeros(handle.data.shape[0], dtype=np.float32)
            for i in range(handle.extended_header.shape[0]):
                angle[i] = handle.extended_header[i]["Alpha tilt"]

            position = np.zeros(shape=(handle.data.shape[0], 3), dtype=np.float32)
            drift = np.zeros(shape=(handle.data.shape[0], 2), dtype=np.float32)
            defocus = np.zeros(shape=(handle.data.shape[0]), dtype=np.float32)
            for i in range(handle.extended_header.shape[0]):

                # Read the positions
                position[i, 0] = handle.extended_header[i]["Shift X"]
                position[i, 1] = handle.extended_header[i]["Shift Y"]
                position[i, 2] = handle.extended_header[i]["Z-Stage"]

                # Read the drift
                drift[i, 0] = handle.extended_header[i]["Shift offset X"]
                drift[i, 1] = handle.extended_header[i]["Shift offset Y"]

                # Read the defocus
                defocus[i] = handle.extended_header[i]["Defocus"]
        else:
            angle = np.zeros(handle.data.shape[0], dtype=np.float32)
            position = np.zeros(shape=(handle.data.shape[0], 3), dtype=np.float32)
            drift = np.zeros(shape=(handle.data.shape[0], 2), dtype=np.float32)
            defocus = np.zeros(handle.data.shape[0], dtype=np.float32)

        # Get the pixel size
        pixel_size = handle.voxel_size["x"]

        # Create the reader
        return Reader(
            handle,
            handle.data,
            angle,
            position,
            pixel_size,
            drift=drift,
            defocus=defocus,
        )

    @classmethod
    def from_nexus(Class, filename):
        """
        Read the simulated data from a nexus file

        Args:
            filename (str): The input filename

        """

        # Read the data from disk
        handle = h5py.File(filename, "r")

        # Get the entry
        entry = handle["entry"]
        assert entry.attrs["NX_class"] == "NXentry"
        definition = entry["definition"][()]
        if isinstance(definition, bytes):
            definition = definition.decode("utf-8")
        assert definition == "NXtomo"

        # Get the data and detector
        data = entry["data"]
        detector = entry["instrument"]["detector"]

        # Get the positions
        position = np.array(
            (data["x_translation"], data["y_translation"], data["z_translation"])
        ).T

        # Get the drifts
        drift = np.array((data["x_drift"], data["y_drift"])).T

        # Get the defocus
        defocus = data.get("defocus", None)

        # Get the pixel size
        pixel_size = detector["x_pixel_size"][0]

        # Create the reader
        return Reader(
            handle,
            data["data"],
            data["rotation_angle"],
            position,
            pixel_size,
            drift=drift,
            defocus=defocus,
        )

    @classmethod
    def from_file(Class, filename):
        """
        Read the simulated data from file

        Args:
            filename (str): The output filename

        """
        extension = os.path.splitext(filename)[1].lower()
        if extension in [".mrc"]:
            return Class.from_mrcfile(filename)
        elif extension in [".h5", ".hdf5", ".nx", ".nxs", ".nexus", "nxtomo"]:
            return Class.from_nexus(filename)
        else:
            raise RuntimeError(f"File with unknown extension: {filename}")


def new(filename, shape=None, pixel_size=1, dtype="float32", vmin=None, vmax=None):
    """
    Create a new file for writing

    Args:
        filename (str): The output filename
        shape (tuple): The output shape
        pixel_size (tuple): The pixel size
        dtype (object): The data type (only used with NexusWriter)
        vmin (int): The minimum value (only used in ImageWriter)
        vmax (int): The maximum value (only used in ImageWriter)

    Returns:
        object: The file writer

    """
    extension = os.path.splitext(filename)[1].lower()
    if extension in [".mrc"]:
        return MrcFileWriter(filename, shape, pixel_size, dtype)
    elif extension in [".h5", ".hdf5", ".nx", ".nxs", ".nexus", "nxtomo"]:
        return NexusWriter(filename, shape, pixel_size, dtype)
    elif extension in [".png", ".jpg", ".jpeg", ".tif", ".tiff"]:
        return ImageWriter(filename, shape, vmin, vmax)
    else:
        raise RuntimeError(f"File with unknown extension: {filename}")


def open(filename):
    """
    Read the simulated data from file

    Args:
        filename (str): The output filename

    """
    return Reader.from_file(filename)



#
# parakeet.data.__init__.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import os.path


def get_path(name):
    """
    Return the path to the data file

    Args:
        name (str): The relative filename

    Returns:
        str: The absolute filename

    """
    return os.path.join(os.path.dirname(__file__), "files", name)


def get_pdb(name):
    """
    Return the path to the data file

    Args:
        name (str): The pdb id

    Returns:
        str: The absolute filename

    """

    def get_filenames_to_check(name):
        return [get_path("%s.%s" % (name, ext)) for ext in ["pdb", "cif"]]

    def get_valid_filenames(name):
        return [
            filename
            for filename in get_filenames_to_check(name)
            if os.path.exists(filename)
        ]

    return get_valid_filenames(name)[0]


def get_4v1w():
    """
    Get the path to the file containing the structure of Apoferratin (4v1w)

    Returns:
        str: The abolsute path to 4v1w

    """
    return get_path("4v1w.cif")


def get_4v5d():
    """
    Get the path to the file containing the structure of ribosome (4v5d)

    Returns:
        str: The abolsute path to 4v5d

    """
    return get_path("4v5d.cif")


def get_6qt9():
    """
    Get the path to the file containing the structure of the icosohedral virus
    SH1 (6qt9)

    Returns:
        str: The abolsute path to 6qt9

    """
    return get_path("6qt9.cif")



#
# parakeet.analyse.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import numpy as np
import maptools
import mrcfile
import os.path
import guanaco
import random
import h5py
import scipy.ndimage
import scipy.spatial.transform
import parakeet.sample
from math import sqrt, ceil

# Set the random seed
random.seed(0)


def reconstruct(image_filename, rec_filename, microscope, simulation, device="gpu"):
    """
    Reconstruct the volume and use 3D CTF correction beforehand if the input image is uncorrected

    """

    # Ensure mrc file
    assert os.path.splitext(image_filename)[1] == ".mrc"

    # Set the corrected filename
    corrected_filename = os.path.join(os.path.dirname(rec_filename), "CORRECTED.dat")

    # Get the parameters for the CTF correction
    nx = microscope.detector.nx
    pixel_size = microscope.detector.pixel_size
    energy = microscope.beam.energy
    defocus = -microscope.lens.c_10
    num_defocus = int((nx * pixel_size) / 100)

    # Set the spherical aberration
    if simulation["inelastic_model"] == "cc_corrected":
        print("Setting spherical aberration to zero")
        spherical_aberration = 0
    else:
        spherical_aberration = microscope.lens.c_30

    astigmatism = microscope.lens.c_12
    astigmatism_angle = microscope.lens.phi_12
    phase_shift = 0

    # Do the reconstruction
    guanaco.reconstruct_file(
        input_filename=image_filename,
        output_filename=rec_filename,
        corrected_filename=corrected_filename,
        centre=None,
        energy=energy,
        defocus=defocus,
        num_defocus=num_defocus,
        spherical_aberration=spherical_aberration,
        astigmatism=astigmatism,
        astigmatism_angle=astigmatism_angle,
        phase_shift=phase_shift,
        angular_weights=True,
        device=device,
    )


def correct(
    image_filename,
    corrected_filename,
    microscope,
    simulation,
    num_defocus=None,
    device="gpu",
):
    """
    Correct the images using 3D CTF correction

    """

    # Ensure mrc file
    assert os.path.splitext(image_filename)[1] == ".mrc"

    # Get the parameters for the CTF correction
    nx = microscope.detector.nx
    pixel_size = microscope.detector.pixel_size
    energy = microscope.beam.energy
    defocus = -microscope.lens.c_10
    if num_defocus is None:
        num_defocus = int((nx * pixel_size) / 100)

    # Set the spherical aberration
    if simulation["inelastic_model"] == "cc_corrected":
        print("Setting spherical aberration to zero")
        spherical_aberration = 0
    else:
        spherical_aberration = microscope.lens.c_30

    astigmatism = microscope.lens.c_12
    astigmatism_angle = microscope.lens.phi_12
    phase_shift = 0

    # Do the reconstruction
    guanaco.correct_file(
        input_filename=image_filename,
        output_filename=corrected_filename,
        centre=None,
        energy=energy,
        defocus=defocus,
        num_defocus=num_defocus,
        spherical_aberration=spherical_aberration,
        astigmatism=astigmatism,
        astigmatism_angle=astigmatism_angle,
        phase_shift=phase_shift,
        device=device,
    )


def average_particles(
    scan,
    sample_filename,
    rec_filename,
    half_1_filename,
    half_2_filename,
    particle_size=0,
    num_particles=0,
):
    """
    Average particles to compute averaged reconstruction

    """

    def rotate_array(data, rotation, offset):

        # Create the pixel indices
        az = np.arange(data.shape[0])
        ay = np.arange(data.shape[1])
        ax = np.arange(data.shape[2])
        x, y, z = np.meshgrid(az, ay, ax, indexing="ij")

        # Create a stack of coordinates
        xyz = np.vstack(
            [
                x.reshape(-1) - offset[0],
                y.reshape(-1) - offset[1],
                z.reshape(-1) - offset[2],
            ]
        ).T

        # create transformation matrix
        r = scipy.spatial.transform.Rotation.from_rotvec(rotation)

        # apply transformation
        transformed_xyz = r.apply(xyz)

        # extract coordinates
        x = transformed_xyz[:, 0] + offset[0]
        y = transformed_xyz[:, 1] + offset[1]
        z = transformed_xyz[:, 2] + offset[2]

        # Reshape
        x = x.reshape(data.shape)
        y = y.reshape(data.shape)
        z = z.reshape(data.shape)

        # sample
        result = scipy.ndimage.map_coordinates(data, [x, y, z], order=1)
        return result

    # Load the sample
    sample = parakeet.sample.load(sample_filename)

    # Get the sample centre
    centre = np.array(sample.centre)

    # Read the reconstruction file
    tomo_file = mrcfile.mmap(rec_filename)
    tomogram = tomo_file.data

    # Get the size of the volume
    voxel_size = np.array(
        (
            tomo_file.voxel_size["x"],
            tomo_file.voxel_size["y"],
            tomo_file.voxel_size["z"],
        )
    )
    size = np.array(tomogram.shape)[[2, 0, 1]] * voxel_size

    # Loop through the
    assert sample.number_of_molecules == 1
    for name, (atoms, positions, orientations) in sample.iter_molecules():

        # Compute the box size based on the size of the particle so that any
        # orientation should fit within the box
        xmin = atoms.data["x"].min()
        xmax = atoms.data["x"].max()
        ymin = atoms.data["y"].min()
        ymax = atoms.data["y"].max()
        zmin = atoms.data["z"].min()
        zmax = atoms.data["z"].max()
        xc = (xmax + xmin) / 2.0
        yc = (ymax + ymin) / 2.0
        zc = (zmax + zmin) / 2.0

        if particle_size == 0:
            half_length = (
                int(ceil(sqrt((xmin - xc) ** 2 + (ymin - yc) ** 2 + (zmin - zc) ** 2)))
                + 1
            )
        else:
            half_length = particle_size // 2
        length = 2 * half_length
        assert len(positions) == len(orientations)
        if num_particles <= 0:
            num_particles = len(positions)
        print(
            "Averaging %d %s particles with box size %d" % (num_particles, name, length)
        )

        # Create the average array
        half_1 = np.zeros(shape=(length, length, length), dtype="float32")
        half_2 = np.zeros(shape=(length, length, length), dtype="float32")
        num_1 = 0
        num_2 = 0

        # Sort the positions and orientations by y
        positions, orientations = zip(
            *sorted(zip(positions, orientations), key=lambda x: x[0][1])
        )

        # Loop through all the particles
        for i, (position, orientation) in enumerate(zip(positions, orientations)):

            # Compute p within the volume
            # start_position = np.array([0, scan["start_pos"], 0])
            p = position - (centre - size / 2.0)  # - start_position
            p[2] = size[2] - p[2]
            print(
                "Particle %d: position = %s, orientation = %s"
                % (
                    i,
                    "[ %.1f, %.1f, %.1f ]" % tuple(p),
                    "[ %.1f, %.1f, %.1f ]" % tuple(orientation),
                )
            )

            # Set the region to extract
            x0 = np.floor(p).astype("int32") - half_length
            x1 = np.floor(p).astype("int32") + half_length
            offset = p - np.floor(p).astype("int32")

            # Get the sub tomogram
            print("Getting sub tomogram")
            sub_tomo = tomogram[x0[1] : x1[1], x0[2] : x1[2], x0[0] : x1[0]]
            if sub_tomo.shape == half_1.shape:

                # Set the data to transform
                data = sub_tomo

                # Reorder input vectors
                offset = np.array(data.shape)[::-1] / 2 + offset[[1, 2, 0]]
                rotation = -np.array(orientation)[[1, 2, 0]]
                rotation[1] = -rotation[1]

                # Rotate the data
                print("Rotating volume")
                data = rotate_array(data, rotation, offset)

                # Add the contribution to the average
                if bool(random.getrandbits(1)):
                    half_1 += data
                    num_1 += 1
                else:
                    half_2 += data
                    num_2 += 1

            # Break if we have enough particles
            if num_1 + num_2 >= num_particles:
                break

        # Average the sub tomograms
        print("Averaging half 1 with %d particles" % num_1)
        print("Averaging half 2 with %d particles" % num_2)
        if num_1 > 0:
            half_1 = half_1 / num_1
        if num_2 > 0:
            half_2 = half_2 / num_2

        # from matplotlib import pylab
        # pylab.imshow(average[half_length, :, :])
        # pylab.show()

        # Save the averaged data
        print("Saving half 1 to %s" % half_1_filename)
        handle = mrcfile.new(half_1_filename, overwrite=True)
        handle.set_data(half_1)
        handle.voxel_size = tomo_file.voxel_size
        print("Saving half 2 to %s" % half_2_filename)
        handle = mrcfile.new(half_2_filename, overwrite=True)
        handle.set_data(half_2)
        handle.voxel_size = tomo_file.voxel_size


def average_all_particles(
    scan, sample_filename, rec_filename, average_filename, particle_size=0
):
    """
    Average particles to compute averaged reconstruction

    """

    def rotate_array(data, rotation, offset):

        # Create the pixel indices
        az = np.arange(data.shape[0])
        ay = np.arange(data.shape[1])
        ax = np.arange(data.shape[2])
        x, y, z = np.meshgrid(az, ay, ax, indexing="ij")

        # Create a stack of coordinates
        xyz = np.vstack(
            [
                x.reshape(-1) - offset[0],
                y.reshape(-1) - offset[1],
                z.reshape(-1) - offset[2],
            ]
        ).T

        # create transformation matrix
        r = scipy.spatial.transform.Rotation.from_rotvec(rotation)

        # apply transformation
        transformed_xyz = r.apply(xyz)

        # extract coordinates
        x = transformed_xyz[:, 0] + offset[0]
        y = transformed_xyz[:, 1] + offset[1]
        z = transformed_xyz[:, 2] + offset[2]

        # Reshape
        x = x.reshape(data.shape)
        y = y.reshape(data.shape)
        z = z.reshape(data.shape)

        # sample
        result = scipy.ndimage.map_coordinates(data, [x, y, z], order=1)
        return result

    # Load the sample
    sample = parakeet.sample.load(sample_filename)

    # Get the sample centre
    centre = np.array(sample.centre)

    # Read the reconstruction file
    tomo_file = mrcfile.mmap(rec_filename)
    tomogram = tomo_file.data

    # Get the size of the volume
    voxel_size = np.array(
        (
            tomo_file.voxel_size["x"],
            tomo_file.voxel_size["y"],
            tomo_file.voxel_size["z"],
        )
    )
    size = np.array(tomogram.shape)[[2, 0, 1]] * voxel_size

    # Loop through the
    assert sample.number_of_molecules == 1
    for name, (atoms, positions, orientations) in sample.iter_molecules():

        # Compute the box size based on the size of the particle so that any
        # orientation should fit within the box
        xmin = atoms.data["x"].min()
        xmax = atoms.data["x"].max()
        ymin = atoms.data["y"].min()
        ymax = atoms.data["y"].max()
        zmin = atoms.data["z"].min()
        zmax = atoms.data["z"].max()
        xc = (xmax + xmin) / 2.0
        yc = (ymax + ymin) / 2.0
        zc = (zmax + zmin) / 2.0

        if particle_size == 0:
            half_length = (
                int(ceil(sqrt((xmin - xc) ** 2 + (ymin - yc) ** 2 + (zmin - zc) ** 2)))
                + 1
            )
        else:
            half_length = particle_size // 2
        length = 2 * half_length
        assert len(positions) == len(orientations)
        num_particles = len(positions)
        print(
            "Averaging %d %s particles with box size %d" % (num_particles, name, length)
        )

        # Create the average array
        average = np.zeros(shape=(length, length, length), dtype="float32")
        num = 0

        # Sort the positions and orientations by y
        positions, orientations = zip(
            *sorted(zip(positions, orientations), key=lambda x: x[0][1])
        )

        # Loop through all the particles
        for i, (position, orientation) in enumerate(zip(positions, orientations)):

            # Compute p within the volume
            start_position = np.array([0, scan["start_pos"], 0])
            p = position - (centre - size / 2.0)  # - start_position
            p[2] = size[2] - p[2]
            print(
                "Particle %d: position = %s, orientation = %s"
                % (
                    i,
                    "[ %.1f, %.1f, %.1f ]" % tuple(p),
                    "[ %.1f, %.1f, %.1f ]" % tuple(orientation),
                )
            )

            # Set the region to extract
            x0 = np.floor(p).astype("int32") - half_length
            x1 = np.floor(p).astype("int32") + half_length
            offset = p - np.floor(p).astype("int32")

            # Get the sub tomogram
            print("Getting sub tomogram")
            sub_tomo = tomogram[x0[1] : x1[1], x0[2] : x1[2], x0[0] : x1[0]]
            if sub_tomo.shape == average.shape:

                # Set the data to transform
                data = sub_tomo

                # Reorder input vectors
                offset = np.array(data.shape)[::-1] / 2 + offset[[1, 2, 0]]
                rotation = -np.array(orientation)[[1, 2, 0]]
                rotation[1] = -rotation[1]

                # Rotate the data
                print("Rotating volume")
                data = rotate_array(data, rotation, offset)

                # Add the contribution to the average

                average += data
                num += 1

        # Average the sub tomograms
        print("Averaging map with %d particles" % num)
        average = average / num

        # from matplotlib import pylab
        # pylab.imshow(average[half_length, :, :])
        # pylab.show()

        # Save the averaged data
        print("Saving map to %s" % average_filename)
        handle = mrcfile.new(average_filename, overwrite=True)
        handle.set_data(average)
        handle.voxel_size = tomo_file.voxel_size


def extract_particles(
    scan, sample_filename, rec_filename, extract_filename, particle_size=0
):
    """
    Extract particles for post-processing

    """

    def rotate_array(data, rotation, offset):

        # Create the pixel indices
        az = np.arange(data.shape[0])
        ay = np.arange(data.shape[1])
        ax = np.arange(data.shape[2])
        x, y, z = np.meshgrid(az, ay, ax, indexing="ij")

        # Create a stack of coordinates
        xyz = np.vstack(
            [
                x.reshape(-1) - offset[0],
                y.reshape(-1) - offset[1],
                z.reshape(-1) - offset[2],
            ]
        ).T

        # create transformation matrix
        r = scipy.spatial.transform.Rotation.from_rotvec(rotation)

        # apply transformation
        transformed_xyz = r.apply(xyz)

        # extract coordinates
        x = transformed_xyz[:, 0] + offset[0]
        y = transformed_xyz[:, 1] + offset[1]
        z = transformed_xyz[:, 2] + offset[2]

        # Reshape
        x = x.reshape(data.shape)
        y = y.reshape(data.shape)
        z = z.reshape(data.shape)

        # sample
        result = scipy.ndimage.map_coordinates(data, [x, y, z], order=1)
        return result

    # Load the sample
    sample = parakeet.sample.load(sample_filename)

    # Get the sample centre
    centre = np.array(sample.centre)

    # Read the reconstruction file
    tomo_file = mrcfile.mmap(rec_filename)
    tomogram = tomo_file.data

    # Get the size of the volume
    voxel_size = np.array(
        (
            tomo_file.voxel_size["x"],
            tomo_file.voxel_size["y"],
            tomo_file.voxel_size["z"],
        )
    )
    size = np.array(tomogram.shape)[[2, 0, 1]] * voxel_size

    # Loop through the
    assert sample.number_of_molecules == 1
    for name, (atoms, positions, orientations) in sample.iter_molecules():

        # Compute the box size based on the size of the particle so that any
        # orientation should fit within the box
        xmin = atoms.data["x"].min()
        xmax = atoms.data["x"].max()
        ymin = atoms.data["y"].min()
        ymax = atoms.data["y"].max()
        zmin = atoms.data["z"].min()
        zmax = atoms.data["z"].max()
        xc = (xmax + xmin) / 2.0
        yc = (ymax + ymin) / 2.0
        zc = (zmax + zmin) / 2.0

        if particle_size == 0:
            half_length = (
                int(ceil(sqrt((xmin - xc) ** 2 + (ymin - yc) ** 2 + (zmin - zc) ** 2)))
                + 1
            )
        else:
            half_length = particle_size // 2
        length = 2 * half_length
        assert len(positions) == len(orientations)
        num_particles = len(positions)
        print(
            "Averaging %d %s particles with box size %d" % (num_particles, name, length)
        )

        # Create the average array
        extract_map = []
        particle_instance = np.zeros(shape=(length, length, length), dtype="float32")
        num = 0

        # Sort the positions and orientations by y
        positions, orientations = zip(
            *sorted(zip(positions, orientations), key=lambda x: x[0][1])
        )

        # Loop through all the particles
        for i, (position, orientation) in enumerate(zip(positions, orientations)):

            # Compute p within the volume
            start_position = np.array([0, scan["start_pos"], 0])
            p = position - (centre - size / 2.0)  # - start_position
            p[2] = size[2] - p[2]
            print(
                "Particle %d: position = %s, orientation = %s"
                % (
                    i,
                    "[ %.1f, %.1f, %.1f ]" % tuple(p),
                    "[ %.1f, %.1f, %.1f ]" % tuple(orientation),
                )
            )

            # Set the region to extract
            x0 = np.floor(p).astype("int32") - half_length
            x1 = np.floor(p).astype("int32") + half_length
            offset = p - np.floor(p).astype("int32")

            # Get the sub tomogram
            print("Getting sub tomogram")
            sub_tomo = tomogram[x0[1] : x1[1], x0[2] : x1[2], x0[0] : x1[0]]
            if sub_tomo.shape == particle_instance.shape:

                # Set the data to transform
                data = sub_tomo

                # Reorder input vectors
                offset = np.array(data.shape)[::-1] / 2 + offset[[1, 2, 0]]
                rotation = -np.array(orientation)[[1, 2, 0]]
                rotation[1] = -rotation[1]

                # Rotate the data
                print("Rotating volume")
                data = rotate_array(data, rotation, offset)

                # Add the contribution to the average

                extract_map.append(data)
                num += 1

        # Average the sub tomograms
        print("Extracting %d particles" % num)
        extract_map = np.array(extract_map)

        # from matplotlib import pylab
        # pylab.imshow(average[half_length, :, :])
        # pylab.show()

        # Save the averaged data
        print("Saving extracted particles to %s" % extract_filename)
        handle = h5py.File(extract_filename, "w")
        data_handle = handle.create_dataset("data", extract_map.shape, chunks=True)
        data_handle[:] = extract_map[:]
        handle.close()


def refine(sample_filename, rec_filename):
    """
    Refine the molecule against the map

    """

    # Load the sample
    sample = parakeet.sample.load(sample_filename)

    # Get the molecule name
    assert sample.number_of_molecules == 1
    name, _ = list(sample.iter_molecules())[0]

    # Get the PDB filename
    pdb_filename = parakeet.data.get_pdb(name)

    # Fit the molecule to the map
    maptools.fit(
        rec_filename,
        pdb_filename,
        output_pdb_filename="refined.pdb",
        resolution=3,
        ncycle=10,
        mode="rigid_body",
        log_filename="fit.log",
    )



import numpy as np
from scipy.spatial.transform import Rotation as R
import pandas as pd
import starfile


class PoseSet:
    """
    A description of poses of the system being simulated.

    """

    def __init__(self, orientations: np.ndarray, shifts: np.ndarray):
        """
        Init the class

        Args:
            orientations (np.ndarray): (n, 3, 3) array of rotation matrices describing the pose of the system.
            shifts (np.ndarray): (n, 3) array of how to shift the sample volume (units: A)

        """
        self.orientations = R.from_matrix(orientations)
        self.shifts = shifts

    @property
    def euler_angles(self):
        """
        Euler angle representation of the orientations.

        The Euler angles are intrinsic, right handed rotations around ZYZ.
        This matches the convention used by XMIPP/RELION.

        """
        return self.orientations.as_euler(seq="ZYZ", degrees=True)

    @property
    def axis_angle(self):
        """
        Axis-angle representation of the orientations.

        Magnitude of the rotation vector corresponds to the angle in radians.

        """
        return self.orientations.as_rotvec()

    def __len__(self):
        """
        The number of poses

        """
        n_orientations = len(self.orientations)  # .shape[0]
        n_shifts = self.shifts.shape[0]
        if n_orientations != n_shifts:
            raise RuntimeError(
                "Number of orientations is different to the number of shifts."
            )
        return n_shifts

    def write_star_file(self, output_file):
        """Write a RELION format STAR file containing simulation metadata."""
        relion_eulers = self.orientations.inv().as_euler("ZYZ", degrees=True)
        relion_shifts = -self.shifts
        data = {
            "rlnAngleRot": relion_eulers[:, 0],
            "rlnAngleTilt": relion_eulers[:, 1],
            "rlnAnglePsi": relion_eulers[:, 2],
            "rlnOriginX": relion_shifts[:, 0],
            "rlnOriginY": relion_shifts[:, 1],
            "rlnOriginZ": relion_shifts[:, 2],
        }
        df = pd.DataFrame.from_dict(data)
        starfile.write(df, output_file, overwrite=True)



#
# parakeet.microscope.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import parakeet.beam
import parakeet.detector
import parakeet.lens


class Microscope(object):
    """
    A class to encapsulate a microscope

    """

    def __init__(
        self, model=None, beam=None, lens=None, detector=None, phase_plate=False
    ):
        """
        Initialise the detector

        Args:
            model (str): The microscope model name
            beam (object): The beam object
            lens (object): The lens object
            detector (object): The detector object
            phase_plate (bool): The phase plate

        """
        self.model = model
        self.beam = beam
        self.lens = lens
        self.detector = detector
        self.phase_plate = phase_plate


def new(model=None, beam=None, objective_lens=None, detector=None, phase_plate=False):
    """
    Make a new detector

    Args:
        model (str): The microscope model
        beam (dict): The beam parameters
        objective_lens (dict): The objective lens parameters
        detector (dict): The detector parameters
        phase_plate (bool): The phase plate

    Returns:
        obj: The detector object

    """

    # Construct the basic models from the input
    beam = parakeet.beam.new(**beam)
    lens = parakeet.lens.new(**objective_lens)
    detector = parakeet.detector.new(**detector)

    # Override the parameters for the different microscope models
    if model == "krios":
        beam.energy = 300
        beam.energy_spread = 2.66 * 1e-6
        beam.acceleration_voltage_spread = 0.8 * 1e-6
        lens.c_30 = 2.7
        lens.c_c = 2.7
        lens.current_spread = 0.33 * 1e-6
    elif model == "talos":
        beam.energy = 200
        beam.energy_spread = 2.66 * 1e-6
        beam.acceleration_voltage_spread = 0.8 * 1e-6
        lens.c_30 = 2.7
        lens.c_c = 2.7
        lens.current_spread = 0.33 * 1e-6
    elif model is not None:
        raise RuntimeError("Unknown microscope model")

    # Return the miroscope object
    return Microscope(
        model=model, beam=beam, lens=lens, detector=detector, phase_plate=phase_plate
    )



#
# Copyright (C) 2019 James Parkhurst
#
# This code is distributed under the BSD license.
#
import os
import subprocess
import sys
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext


class CMakeBuild(build_ext):
    """
    Build the extensions

    """

    def build_extensions(self):

        # Set the cmake directory
        cmake_lists_dir = os.path.abspath(".")

        # Ensure the build directory exists
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # Run Cmake once
        ext = self.extensions[0]

        # Get the directory
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # Arguments to cmake
        cmake_args = [
            "-DCMAKE_BUILD_TYPE=%s" % "Release",
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=%s" % extdir,
            "-DPYTHON_EXECUTABLE=%s" % sys.executable,
        ]

        # Config and the extension
        subprocess.check_call(
            ["cmake", cmake_lists_dir] + cmake_args, cwd=self.build_temp
        )

        # Build the extension
        subprocess.check_call(["cmake", "--build", "."], cwd=self.build_temp)


def main():
    """
    Setup the package

    """
    tests_require = ["pytest", "pytest-cov", "mock"]

    setup(
        package_dir={"": "src"},
        packages=find_packages(where="src"),
        setup_requires=["dask", "pytest-runner"],
        install_requires=[
            "distributed",
            "dask_jobqueue",
            "gemmi",
            "guanaco",
            "h5py",
            "maptools",
            "mrcfile",
            "numpy",
            "pandas",
            "pillow",
            "python-multem",
            "scipy",
            "pyyaml",
            "pandas",
            "starfile",
        ],
        tests_require=tests_require,
        test_suite="tests",
        ext_modules=[Extension("parakeet_ext", [])],
        cmdclass={"build_ext": CMakeBuild},
        entry_points={
            "console_scripts": [
                "parakeet.read_pdb=parakeet.command_line:read_pdb",
                "parakeet.export=parakeet.command_line:export",
                "parakeet.config.show=parakeet.command_line.config:show",
                "parakeet.config.edit=parakeet.command_line.config:edit",
                "parakeet.sample.new=parakeet.command_line.sample:new",
                "parakeet.sample.add_molecules=parakeet.command_line.sample:add_molecules",
                "parakeet.sample.mill=parakeet.command_line.sample:mill",
                "parakeet.sample.sputter=parakeet.command_line.sample:sputter",
                "parakeet.sample.show=parakeet.command_line.sample:show",
                "parakeet.simulate.projected_potential=parakeet.command_line.simulate:projected_potential",
                "parakeet.simulate.exit_wave=parakeet.command_line.simulate:exit_wave",
                "parakeet.simulate.optics=parakeet.command_line.simulate:optics",
                "parakeet.simulate.image=parakeet.command_line.simulate:image",
                "parakeet.simulate.simple=parakeet.command_line.simulate:simple",
                "parakeet.simulate.ctf=parakeet.command_line.simulate:ctf",
                "parakeet.analyse.reconstruct=parakeet.command_line.analyse:reconstruct",
                "parakeet.analyse.average_particles=parakeet.command_line.analyse:average_particles",
                "parakeet.analyse.average_all_particles=parakeet.command_line.analyse:average_all_particles",
                "parakeet.analyse.extract_particles=parakeet.command_line.analyse:extract_particles",
                "parakeet.analyse.refine=parakeet.command_line.analyse:refine",
                "parakeet.analyse.correct=parakeet.command_line.analyse:correct",
            ]
        },
        extras_require={
            "build_sphinx": ["sphinx", "sphinx_rtd_theme"],
            "test": tests_require,
        },
        include_package_data=True,
    )


if __name__ == "__main__":
    main()



#
# parakeet.scan.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
from abc import ABC, abstractmethod

import numpy as np
from math import pi
from scipy.stats import special_ortho_group
from scipy.spatial.transform import Rotation as R

from parakeet.pose import PoseSet


class Scan(ABC):
    """
    A base class defining the interface for image sampling in a simulation.

    """

    @property
    @abstractmethod
    def poses(self) -> PoseSet:
        pass

    @property
    @abstractmethod
    def exposure_times(self) -> np.ndarray:
        pass

    @property
    def is_single_axis_scan(self):
        return False

    @property
    def is_uniform_angular_scan(self):
        return False

    def __len__(self) -> int:
        """
        The number of images to sample

        """
        return len(self.poses)


class SingleAxisScan(Scan):
    """
    A scan of angles around a single rotation axis.

    """

    def __init__(self, axis=None, angles=None, positions=None, exposure_time=None):
        """
        Initialise the scan

        Args:
            axis (tuple): The rotation axis
            angles (list): The rotation angles (units: degrees)
            positions (list): The positions to shift (units: A)
            exposure_time (float): The exposure time (units: seconds)

        """
        if axis is None:
            self.axis = np.array((0, 1, 0))
        else:
            self.axis = np.array(axis)
        if angles is None:
            self.angles = np.array([0])
        else:
            self.angles = np.array(angles)
        if positions is None:
            self.positions = np.zeros(shape=(len(angles), 3), dtype=np.float32)
        else:
            self.positions = np.array([self.axis * p for p in positions])
        assert len(self.angles) == len(self.positions)
        self.exposure_time = exposure_time

    @property
    def poses(self) -> PoseSet:
        axis = np.array(self.axis)
        axis = axis / np.linalg.norm(axis)
        angles = np.array(self.angles) * pi / 180.0
        orientations = np.array([axis * a for a in angles])
        orientations = R.from_rotvec(orientations).as_matrix()
        shifts = np.array([axis * p for p in self.positions])
        return PoseSet(orientations, axis * shifts)

    @property
    def exposure_times(self):
        """
        Returns:
            np.ndarray: The exposure times array

        """
        return np.ones(len(self)) * self.exposure_time

    @property
    def is_single_axis_scan(self):
        return True


class UniformAngularScan(Scan):
    """
    A uniform scan of orientations, no shifts.

    """

    def __init__(self, n: int, metadata_file: str = None):
        """
        Args:
            n (int): The number of uniform orientational samples

        """
        self.n = int(n)
        self.metadata_file = metadata_file

        # Draw n uniform samples from SO(3)
        self.orientations = R.from_matrix(special_ortho_group.rvs(dim=3, size=self.n))

    @property
    def exposure_time(self) -> float:
        return 1

    @property
    def angles(self) -> np.ndarray:
        return np.linalg.norm(self.orientations.as_rotvec(), axis=1) * 180.0 / pi

    @property
    def positions(self):
        return np.zeros(shape=(self.n, 3), dtype=np.float32)

    @property
    def exposure_times(self) -> np.ndarray:
        return np.ones(len(self)) * self.exposure_time

    @property
    def poses(self) -> PoseSet:
        # Create and return PoseSet object
        return PoseSet(self.orientations.as_matrix(), self.positions)

    @property
    def is_uniform_angular_scan(self):
        return True


def new(
    mode="still",
    axis=(0, 1, 0),
    angles=None,
    positions=None,
    start_angle=0,
    step_angle=0,
    start_pos=0,
    step_pos=0,
    num_images=1,
    exposure_time=1,
    metadata_file=None,
):
    """
    Create an scan

    If angles or positions is None they are generated form the other
    parameters.

    Args:
        mode (str): The type of scan (still, tilt_series, dose_symmetric, helical_scan)
        axis (array): The rotation axis
        angles (array): The rotation angles
        positions (array): The positions
        start_angle (float): The starting angle (deg)
        step_angle (float): The angle step (deg)
        start_pos (float): The starting position (A)
        step_pos (float): The step in position (A)
        num_images (int): The number of images
        exposure_time (float): The exposure time (seconds)
        metadata_file (str): file in which to save experimental metadata
                             !! only for single particle !!

    Returns:
        object: The scan object

    """
    if mode == "single_particle":
        return UniformAngularScan(n=num_images, metadata_file=metadata_file)
    if angles is None:
        if mode == "still":
            angles = [start_angle]
        elif mode == "tilt_series":
            angles = start_angle + step_angle * np.arange(num_images)
        elif mode == "dose_symmetric":
            angles = start_angle + step_angle * np.arange(num_images)
            angles = np.array(sorted(angles, key=lambda x: abs(x)))
        elif mode == "helical_scan":
            angles = start_angle + step_angle * np.arange(num_images)
        else:
            raise RuntimeError(f"Scan mode not recognised: {mode}")
    if positions is None:
        if mode == "still":
            positions = [start_pos]
        elif mode == "tilt_series":
            positions = np.full(
                shape=len(angles), fill_value=start_pos, dtype=np.float32
            )
        elif mode == "dose_symmetric":
            positions = np.full(
                shape=len(angles), fill_value=start_pos, dtype=np.float32
            )
        elif mode == "helical_scan":
            positions = start_pos + step_pos * np.arange(num_images)
        else:
            raise RuntimeError(f"Scan mode not recognised: {mode}")
    return SingleAxisScan(
        axis=axis, angles=angles, positions=positions, exposure_time=exposure_time
    )



#
# parakeet.futures.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import parakeet.config


def as_completed(futures):
    import dask.distributed

    return dask.distributed.as_completed(futures)


def factory(method="sge", max_workers=1):
    """
    Configure the future to use for parallel processing

    Args:
        method (str): The cluster method (sge)
        max_workers (int): The number of worker processes

    """
    import dask_jobqueue
    import dask.distributed

    if method == "sge":

        # Create the SGECluster object
        # For each worker:
        #   - request 1 gpu and 1 core
        #   - use 1 process per worker and 1 thread per process
        # This ensures that only 1 job will be run on each worker at any time
        # which is required to ensure that there is not any competition for
        # gpu resources (by default dask will try to run many jobs at the same
        # time which causes errors and instability).
        cluster = dask_jobqueue.SGECluster(
            cores=1,
            processes=1,
            nthreads=1,
            memory="64 GB",
            queue="all.q",
            project="tomography",
            resource_spec="gpu=1",
            job_extra=["-V"],
            name="parakeet",
            local_directory=parakeet.config.temp_directory(),
            log_directory=parakeet.config.temp_directory(),
        )

        # Set the number of worker nodes
        cluster.scale(max_workers)

        # Return the client
        executor = dask.distributed.Client(cluster)
    else:
        raise RuntimeError(f"Unknown multiprocessing method: {method}")

    # Return the executor
    return executor



#
# parakeet.command_line.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import argparse
import gemmi
import logging
import logging.config
import numpy
import random
import parakeet.io
import parakeet.config
import parakeet.sample

# Get the logger
logger = logging.getLogger(__name__)


def configure_logging():
    """
    Configure the logging

    """

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "handlers": {
                "stream": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                }
            },
            "loggers": {
                "parakeet": {
                    "handlers": ["stream"],
                    "level": "DEBUG",
                    "propagate": True,
                }
            },
        }
    )


def rebin(data, shape):
    """
    Rebin a multidimensional array

    Args:
        data (array): The input array
        shape (tuple): The new shape

    """
    f = numpy.fft.fft2(data)
    f = numpy.fft.fftshift(f)
    yc, xc = data.shape[0] // 2, data.shape[1] // 2
    yh = shape[0] // 2
    xh = shape[1] // 2
    x0 = xc - xh
    y0 = yc - yh
    x1 = x0 + shape[1]
    y1 = y0 + shape[0]
    y, x = numpy.mgrid[0 : data.shape[0], 0 : data.shape[1]]
    r = (y - yc) ** 2 / yh**2 + (x - xc) ** 2 / xh**2
    mask = r < 1.0
    f = f * mask
    f = f[y0:y1, x0:x1]
    f = numpy.fft.ifftshift(f)
    d = numpy.fft.ifft2(f)
    return d.real


def filter_image(data, pixel_size, resolution, shape):
    """
    Filter the image

    Args:
        data (array): The input array
        pixel_size (float): The pixel size (A)
        resolution (float): The filter resolution
        shape (str): The filter shape

    """
    if pixel_size == 0:
        pixel_size = 1
    f = numpy.fft.fft2(data)
    f = numpy.fft.fftshift(f)
    yc, xc = data.shape[0] // 2, data.shape[1] // 2
    y, x = numpy.mgrid[0 : data.shape[0], 0 : data.shape[1]]
    r = (
        numpy.sqrt(
            (y - yc) ** 2 / data.shape[0] ** 2 + (x - xc) ** 2 / data.shape[1] ** 2
        )
        / pixel_size
    )
    if shape == "square":
        g = r < 1.0 / resolution
    elif shape == "guassian":
        g = numpy.exp(-0.5 * r**2 * resolution**2)
    f = f * g
    f = numpy.fft.ifftshift(f)
    d = numpy.fft.ifft2(f)
    return d.real


def export(argv=None):
    """
    Convert the input file type to a different file type

    """
    # Create the argument parser
    parser = argparse.ArgumentParser(description="Read a PDB file")

    # Add an argument for the filename
    parser.add_argument("filename", type=str, default=None, help="The input filename")

    # Add an argument for the filename
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        required=True,
        dest="output",
        help="The output filename",
    )
    parser.add_argument(
        "--rot90",
        type=bool,
        default=False,
        dest="rot90",
        help="Rotate the image 90deg counter clockwise",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--rotation_range",
        type=str,
        default=None,
        dest="rotation_range",
        help="Select a rotation range",
    )
    group.add_argument(
        "--select_images",
        type=str,
        default=None,
        dest="select_images",
        help="Select a range of images (start,stop,step)",
    )
    parser.add_argument(
        "--roi", type=str, default=None, dest="roi", help="Select a region of interest"
    )
    parser.add_argument(
        "--complex_mode",
        choices=[
            "complex",
            "real",
            "imaginary",
            "amplitude",
            "phase",
            "phase_unwrap",
            "square",
            "imaginary_square",
        ],
        default="complex",
        dest="complex_mode",
        help="How to treat complex numbers",
    )
    parser.add_argument(
        "--interlace",
        type=int,
        default=None,
        dest="interlace",
        help="Interlace the scan",
    )
    parser.add_argument(
        "--rebin", type=int, default=1, dest="rebin", help="The rebinned factor"
    )
    parser.add_argument(
        "--filter_resolution",
        dest="filter_resolution",
        type=float,
        default=None,
        help="The resolution",
    )
    parser.add_argument(
        "--filter_shape",
        dest="filter_shape",
        type=str,
        choices=["square", "gaussian"],
        default=None,
        help="The shape of the filter",
    )
    parser.add_argument(
        "--vmin",
        type=float,
        default=None,
        dest="vmin",
        help="The minimum pixel value when exporting to an image",
    )
    parser.add_argument(
        "--vmax",
        type=float,
        default=None,
        dest="vmax",
        help="The maximum pixel value when exporting to an image",
    )
    parser.add_argument(
        "--sort",
        type=str,
        default=None,
        dest="sort",
        choices=["angle"],
        help="Sort the images",
    )

    # Parse the arguments
    args = parser.parse_args(argv)

    # Configure some basic logging
    configure_logging()

    # Read the input
    logger.info(f"Reading data from {args.filename}")
    reader = parakeet.io.open(args.filename)

    # Get the shape and indices to read
    if args.select_images is not None:
        logger.info("Selecting image range %s" % args.select_images)
        item = tuple(map(int, args.select_images.split(",")))
        indices = list(range(*item))
    else:
        indices = list(range(reader.shape[0]))

    # Get the shape and indices to read
    if args.rotation_range is not None:
        args.rotation_range = args.rotation_range.split(";")
        indices = []
        for p in range(0, len(args.rotation_range)):
            current_rotation_range = tuple(map(int, args.rotation_range[p].split(",")))
            for i in range(reader.shape[0]):
                angle = reader.angle[i]
                if (
                    angle >= current_rotation_range[0]
                    and angle < current_rotation_range[1]
                ):
                    indices.append(i)
                    logger.info(f"    Image {i} added as within the rotation range(s)")
        indices = list(set(indices))
        indices.sort()

    # Interlace the images
    if args.interlace is not None:
        if args.interlace >= 1:
            interlaced_indices = []
            for i in range(args.interlace):
                interlaced_indices.extend(
                    [indices[j] for j in range(i, len(indices), args.interlace)]
                )
            indices = interlaced_indices
        else:
            random.shuffle(indices)

    # Sort the images
    if args.sort is not None:
        if args.sort == "angle":
            indices = sorted(indices, key=lambda i: reader.angle[indices[i]])

    # Get the region of interest
    if args.roi is not None:
        x0, y0, x1, y1 = tuple(map(int, args.roi.split(",")))
        assert x1 > x0
        assert y1 > y0
    else:
        x0, y0, x1, y1 = 0, 0, reader.data.shape[2], reader.data.shape[1]

    # If squared and dtype is complex then change to float
    if args.complex_mode != "complex":
        dtype = "float64"
    else:
        dtype = reader.data.dtype.name

    # Set the dataset shape
    shape = (len(indices), y1 - y0, x1 - x0)

    # If rotating, then rotate shape
    if args.rot90:
        shape = (shape[0], shape[2], shape[1])

    # Rebin
    if args.rebin != 1:
        shape = (shape[0], shape[1] // args.rebin, shape[2] // args.rebin)
        pixel_size = reader.pixel_size * args.rebin
    else:
        pixel_size = reader.pixel_size

    # Create the write
    logger.info(f"Writing data to {args.output}")
    writer = parakeet.io.new(
        args.output, shape=shape, pixel_size=pixel_size, dtype=dtype
    )

    # If converting to images, determine min and max
    if writer.is_image_writer:
        if args.vmin is None or args.vmax is None:
            logger.info("Computing min and max of dataset:")
            min_image = []
            max_image = []
            for i in indices:

                # Transform if necessary
                image = {
                    "complex": lambda x: x,
                    "real": lambda x: numpy.real(x),
                    "imaginary": lambda x: numpy.imag(x),
                    "amplitude": lambda x: numpy.abs(x),
                    "phase": lambda x: numpy.real(numpy.angle(x)),
                    "phase_unwrap": lambda x: numpy.unwrap(numpy.real(numpy.angle(x))),
                    "square": lambda x: numpy.abs(x) ** 2,
                    "imaginary_square": lambda x: numpy.imag(x) ** 2 + 1,
                }[args.complex_mode](reader.data[i, y0:y1, x0:x1])

                min_image.append(numpy.min(image))
                max_image.append(numpy.max(image))
                logger.info(
                    "    Reading image %d: min/max: %.2f/%.2f"
                    % (i, min_image[-1], max_image[-1])
                )
            writer.vmin = min(min_image)
            writer.vmax = max(max_image)
            logger.info("Min: %f" % writer.vmin)
            logger.info("Max: %f" % writer.vmax)
        if args.vmin:
            writer.vmin = args.vmin
        if args.vmax:
            writer.vmax = args.vmax

    # Write the data
    for j, i in enumerate(indices):
        logger.info(f"    Copying image {i} -> image {j}")

        # Get the image info
        image = reader.data[i, y0:y1, x0:x1]
        angle = reader.angle[i]
        position = reader.position[i]
        pixel_size = reader.pixel_size
        drift = reader.drift[i]
        defocus = reader.defocus[i]

        # Rotate if necessary
        if args.rot90:
            image = numpy.rot90(image)
            position = (position[1], position[0], position[2])

        # Transform if necessary
        image = {
            "complex": lambda x: x,
            "real": lambda x: numpy.real(x),
            "imaginary": lambda x: numpy.imag(x),
            "amplitude": lambda x: numpy.abs(x),
            "phase": lambda x: numpy.real(numpy.angle(x)),
            "phase_unwrap": lambda x: numpy.unwrap(numpy.real(numpy.angle(x))),
            "square": lambda x: numpy.abs(x) ** 2,
            "imaginary_square": lambda x: numpy.imag(x) ** 2 + 1,
        }[args.complex_mode](image)

        # Filter the images
        if args.filter_shape is not None:
            image = filter_image(
                image, pixel_size, args.filter_resolution, args.filter_shape
            )

        # Rebin the array
        if args.rebin != 1:
            new_shape = numpy.array(image.shape) // args.rebin
            image = rebin(image, new_shape)

        # Write the image info
        writer.data[j, :, :] = image
        writer.angle[j] = angle
        writer.position[j] = position
        if drift is not None:
            writer.drift[j] = drift
        if defocus is not None:
            writer.defocus[j] = defocus

    # Update the writer
    writer.update()


def read_pdb():
    """
    Read the given PDB file and show the atom positions

    """

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Read a PDB file")

    # Add an argument for the filename
    parser.add_argument(
        "-f",
        "--filename",
        type=str,
        default=None,
        dest="filename",
        help="The path to the PDB file",
    )

    # Parse the arguments
    args = parser.parse_args()

    # Check a filename has been given
    if args.filename is None:
        parser.print_help()
        exit(0)

    # Configure some basic logging
    configure_logging()

    # Read the structure
    structure = gemmi.read_structure(args.filename)

    # Iterate through atoms
    prefix = " " * 4
    logger.info("Structure: %s" % structure.name)
    for model in structure:
        logger.info("%sModel: %s" % (prefix, model.name))
        for chain in model:
            logger.info("%sChain: %s" % (prefix * 2, chain.name))
            for residue in chain:
                logger.info("%sResidue: %s" % (prefix * 3, residue.name))
                for atom in residue:
                    logger.info(
                        "%sAtom: %s, %f, %f, %f, %f, %f, %f"
                        % (
                            prefix * 4,
                            atom.element.name,
                            atom.pos.x,
                            atom.pos.y,
                            atom.pos.z,
                            atom.occ,
                            atom.charge,
                            parakeet.sample.get_atom_sigma(atom),
                        )
                    )
